import argparse
import textwrap

import configargparse

from falcon import __version__


class NewlineTextHelpFormatter(argparse.HelpFormatter):

    def _fill_text(self, text, width, indent):
        return '\n'.join(
            textwrap.fill(line, width, initial_indent=indent,
                          subsequent_indent=indent,
                          replace_whitespace=False).strip()
            for line in text.splitlines(keepends=True))


class Config:
    """
    Commandline and file-based configuration.

    Configuration settings can be specified in a config.ini file (by default in
    the working directory), or as command-line arguments.
    """

    def __init__(self) -> None:
        """
        Initialize the configuration settings and provide sensible default
        values if possible.
        """

        self._parser = configargparse.ArgParser(
            description=f'falcon: Fast spectrum clustering using nearest '
                        f'neighbor searching\n'
                        f'==============================================='
                        f'==================\n\n'
                        f'falcon version {__version__}\n\n'
                        f'Official code website: '
                        f'https://github.com/bittremieux/falcon\n\n',
            default_config_files=['config.ini'],
            args_for_setting_config_path=['-c', '--config'],
            formatter_class=NewlineTextHelpFormatter)

        # IO
        self._parser.add_argument(
            'input_filenames', nargs='+',
            help='Input peak files (supported formats: .mzML, .mzXML, .MGF).')
        self._parser.add_argument(
            'output_filename', help='Output file name.')
        self._parser.add_argument(
            '--work_dir', default=None,
            help='Working directory (default: temporary directory).')
        self._parser.add_argument(
            '--overwrite', action='store_true',
            help="Overwrite existing results (default: don't overwrite).")
        self._parser.add_argument(
            '--export_representatives', action='store_true',
            help='Export cluster representatives to an MGF file '
                 '(default: no export).')
        self._parser.add_argument(
            '--export_include_singletons', action='store_true',
            help="Include singletons in the cluster representatives MGF file "
                 "(default: don't include singletons).")
        self._parser.add_argument(
            '--usi_pxd', default='USI000000',
            help='ProteomeXchange dataset identifier to create Universal '
                 'Spectrum Identifier references (default: %(default)s).')

        # CLUSTERING
        self._parser.add_argument(
            '--precursor_tol', nargs=2, default=[20, 'ppm'],
            help='Precursor tolerance mass and mode (default: 20 ppm). '
                 'Mode should be either "ppm" or "Da".')
        self._parser.add_argument(
            '--rt_tol', type=float, default=None,
            help='Retention time tolerance (default: no retention time '
                 'filtering).')
        self._parser.add_argument(
            '--fragment_tol', type=float, default=0.05,
            help='Fragment mass tolerance in m/z (default: %(default)s m/z).')

        self._parser.add_argument(
            '--eps', type=float, default=0.1,
            help='The eps parameter (cosine distance) for DBSCAN clustering '
                 '(default: %(default)s). Relevant cosine distance thresholds '
                 'are typically between 0.05 and 0.30.')
        self._parser.add_argument(
            '--min_samples', type=int, default=2,
            help='The min_samples parameter for DBSCAN clustering (default: '
                 '%(default)s).')

        self._parser.add_argument(
            '--mz_interval', type=int, default=1,
            help='Precursor m/z interval (centered around x.5 Da) to process '
                 'spectra simultaneously (default: %(default)s m/z).')
        self._parser.add_argument(
            '--hash_len', default=800, type=int,
            help='Hashed vector length (default: %(default)s).')
        self._parser.add_argument(
            '--n_neighbors', default=64, type=int,
            help='Number of neighbors to include in the pairwise distance '
                 'matrix for each spectrum (default: %(default)s).')
        self._parser.add_argument(
            '--n_neighbors_ann', default=128, type=int,
            help='Number of neighbors to retrieve from the nearest neighbor '
                 'indexes prior to precursor tolerance filtering '
                 '(default: %(default)s).')
        self._parser.add_argument(
            '--batch_size', default=2**16, type=int,
            help='Number of spectra to process simultaneously '
                 '(default: %(default)s).')
        self._parser.add_argument(
            '--n_probe', default=32, type=int,
            help='Maximum number of lists in the inverted index to inspect '
                 'during querying (default: %(default)s).')

        # PREPROCESSING
        self._parser.add_argument(
            '--min_peaks', default=5, type=int,
            help='Discard spectra with fewer than this number of peaks '
                 '(default: %(default)s).')
        self._parser.add_argument(
            '--min_mz_range', default=250., type=float,
            help='Discard spectra with a smaller mass range '
                 '(default: %(default)s m/z).')
        self._parser.add_argument(
            '--min_mz', default=101., type=float,
            help='Minimum peak m/z value (inclusive, '
                 'default: %(default)s m/z).')
        self._parser.add_argument(
            '--max_mz', default=1500., type=float,
            help='Maximum peak m/z value (inclusive, '
                 'default: %(default)s m/z).')
        self._parser.add_argument(
            '--remove_precursor_tol', default=1.5, type=float,
            help='Window around the precursor mass to remove peaks '
                 '(default: %(default)s m/z).')
        self._parser.add_argument(
            '--min_intensity', default=0.01, type=float,
            help='Remove peaks with a lower intensity relative to the base '
                 'intensity (default: %(default)s).')
        self._parser.add_argument(
            '--max_peaks_used', default=50, type=int,
            help='Only use the specified most intense peaks in the spectra '
                 '(default: %(default)s).')
        self._parser.add_argument(
            '--scaling', default='off', type=str,
            choices=['off', 'root', 'log', 'rank'],
            help='Peak scaling method used to reduce the influence of very '
                 'intense peaks (default: %(default)s).')

        # Filled in 'parse', contains the specified settings.
        self._namespace = None

    def parse(self, args_str: str = None) -> None:
        """
        Parse the configuration settings.

        Parameters
        ----------
        args_str : str
            If None, the arguments are taken from sys.argv. Arguments that are
            not explicitly specified are taken from the configuration file.
        """
        self._namespace = vars(self._parser.parse_args(args_str))

        self._namespace['precursor_tol'][0] = \
            float(self._namespace['precursor_tol'][0])

    def __getattr__(self, option):
        if self._namespace is None:
            raise RuntimeError('The configuration has not been initialized')
        return self._namespace[option]

    def __getitem__(self, item):
        return self.__getattr__(item)


config = Config()



import logging
import math
import pickle
import tempfile
from typing import Callable, Iterator, List, Optional, Tuple

import faiss
import fastcluster
import joblib
import numba as nb
import numpy as np
import pandas as pd
import scipy.sparse as ss
import tqdm
from scipy.cluster.hierarchy import fcluster
from scipy.spatial.distance import squareform
# noinspection PyProtectedMember
from sklearn.cluster._dbscan_inner import dbscan_inner
from sklearn.metrics import pairwise_distances


logger = logging.getLogger('falcon')


def compute_pairwise_distances(
        n_spectra: int, bucket_filenames: List[str],
        process_spectrum: Callable, vectorize: Callable,
        precursor_tol_mass: float, precursor_tol_mode: str, rt_tol: float,
        n_neighbors: int, n_neighbors_ann: int, batch_size: int, n_probe: int)\
        -> Tuple[ss.csr_matrix, pd.DataFrame]:
    """
    Compute a pairwise distance matrix for the persisted spectra with the given
    precursor charge.

    Parameters
    ----------
    n_spectra: int
        The total number of spectra to be processed.
    bucket_filenames : List[str]
        List of bucket file names.
    process_spectrum : Callable
        Function to preprocess the spectra.
    vectorize : Callable
        Function to convert the spectra to vectors.
    precursor_tol_mass : float
        The precursor tolerance mass for vectors to be considered as neighbors.
    precursor_tol_mode : str
        The unit of the precursor m/z tolerance ('Da' or 'ppm').
    rt_tol : float
        The retention time tolerance for vectors to be considered as neighbors.
        If `None`, do not filter neighbors on retention time.
    n_neighbors : int
        The final (maximum) number of neighbors to retrieve for each vector.
    n_neighbors_ann : int
        The number of neighbors to retrieve using the ANN index. This can
        exceed the final number of neighbors (`n_neighbors`) to maximize the
        number of neighbors within the precursor m/z tolerance.
    batch_size : int
        The number of vectors to be simultaneously processed.
    n_probe : int
        The number of cells to visit during ANN querying.

    Returns
    -------
    Tuple[ss.csr_matrix, pd.DataFrame]
        A tuple with the sparse pairwise distance matrix containing the cosine
        distances between similar neighbors for the given vectors, and the
        corresponding spectrum metadata (identifier, precursor charge,
        precursor m/z).
    """
    logger.debug('Compute nearest neighbor pairwise distances (%d spectra, %d'
                 ' neighbors)', n_spectra, n_neighbors)
    max_num_embeddings = n_spectra * n_neighbors
    dtype = (np.int32 if max_num_embeddings < np.iinfo(np.int32).max
             else np.int64)
    distances = np.zeros(max_num_embeddings, np.float32)
    indices = np.zeros(max_num_embeddings, dtype)
    indptr = np.zeros(n_spectra + 1, dtype)
    # Create the ANN indexes (if this hasn't been done yet) and calculate
    # pairwise distances.
    metadata = _build_query_ann_index(
        bucket_filenames, process_spectrum, vectorize, n_probe, batch_size,
        n_neighbors, n_neighbors_ann, precursor_tol_mass, precursor_tol_mode,
        rt_tol, distances, indices, indptr)
    # Update the number of spectra because of skipped low-quality spectra.
    n_spectra = len(metadata)
    indptr = indptr[:n_spectra + 1]
    distances, indices = distances[:indptr[-1]], indices[:indptr[-1]]
    # Convert to a sparse pairwise distance matrix. This matrix might not be
    # entirely symmetrical, but that shouldn't matter too much.
    logger.debug('Construct %d-by-%d sparse pairwise distance matrix with %d '
                 'non-zero values', n_spectra, n_spectra, len(distances))
    pairwise_dist_matrix = ss.csr_matrix(
        (distances, indices, indptr), (n_spectra, n_spectra), np.float32,
        False)
    return pairwise_dist_matrix, metadata


def _build_query_ann_index(
        bucket_filenames: List[str], process_spectrum: Callable,
        vectorize: Callable, n_probe: int, batch_size: int, n_neighbors: int,
        n_neighbors_ann: int, precursor_tol_mass: float,
        precursor_tol_mode: str, rt_tol: float, distances: np.ndarray,
        indices: np.ndarray, indptr: np.ndarray) \
        -> pd.DataFrame:
    """
    Create ANN index(es) for spectra with the given charge per precursor m/z
    split.

    Parameters
    ----------
    bucket_filenames : List[str]
        List of bucket file names.
    process_spectrum : Callable
        Function to preprocess the spectra.
    vectorize : Callable
        Function to convert the spectra to vectors.
    n_probe : int
        The number of cells to consider during NN index querying.
    batch_size : int
        The number of vectors to be simultaneously added to the index.
    n_neighbors : int
        The final (maximum) number of neighbors to retrieve for each vector.
    n_neighbors_ann : int
        The number of neighbors to retrieve using the ANN index. This can
        exceed the final number of neighbors (`n_neighbors`) to maximize the
        number of neighbors within the precursor m/z tolerance.
    precursor_tol_mass : float
        The precursor tolerance mass for vectors to be considered as neighbors.
    precursor_tol_mode : str
        The unit of the precursor m/z tolerance ('Da' or 'ppm').
    rt_tol : float
        The retention time tolerance for vectors to be considered as neighbors.
        If `None`, do not filter neighbors on retention time.
    distances : np.ndarray
        The nearest neighbor distances. See `scipy.sparse.csr_matrix` (`data`).
    indices : np.ndarray
        The column indices for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    indptr : np.ndarray
        The index pointers for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.

    Returns
    -------
    pd.DataFrame
        Metadata (identifier, precursor charge, precursor m/z, retention time)
        of the spectra for which indexes were built.
    """
    identifiers, precursor_mzs, rts = [], [], []
    indptr_i = 0
    # Find neighbors per specified precursor m/z interval.
    for pkl_filename in tqdm.tqdm(bucket_filenames, desc='Buckets queried',
                                  unit='bucket'):
        # Read the spectra for the m/z split.
        spectra_split, precursor_mzs_split, rts_split = [], [], []
        with open(pkl_filename, 'rb') as f_in:
            for spec_raw in pickle.load(f_in):
                spec_processed = process_spectrum(spec_raw)
                # Discard low-quality spectra.
                if spec_processed is not None:
                    spectra_split.append(spec_processed)
                    identifiers.append(spec_processed.identifier)
                    precursor_mzs_split.append(spec_processed.precursor_mz)
                    rts_split.append(spec_processed.retention_time)
        if len(spectra_split) == 0:
            continue
        precursor_mzs.append(np.asarray(precursor_mzs_split))
        rts.append(np.asarray(rts_split))
        # Convert the spectra to vectors.
        vectors_split = vectorize(spectra_split)
        n_split, dim = vectors_split.shape
        # Figure out a decent value for the n_list hyperparameter based on
        # the number of vectors.
        # Rules of thumb from the Faiss wiki:
        # https://github.com/facebookresearch/faiss/wiki/Guidelines-to-choose-an-index#how-big-is-the-dataset
        if n_split == 0:
            continue
        if n_split < 10**2:
            # Use a brute-force index instead of an ANN index when there
            # are only a few items.
            n_list = -1
        elif n_split < 10**6:
            n_list = 2**math.floor(math.log2(n_split / 39))
        elif n_split < 10**7:
            n_list = 2**16
        elif n_split < 10**8:
            n_list = 2**18
        else:
            n_list = 2**20
            if n_split > 10**9:
                logger.warning('More than 1B vectors to be indexed, consider '
                               'decreasing the ANN size')
        # Create an ANN index using the inner product (proxy for cosine
        # distance) for fast NN queries.
        if n_list <= 0:
            index = faiss.IndexIDMap(faiss.IndexFlatIP(dim))
        else:
            index = faiss.IndexIVFFlat(faiss.IndexFlatIP(dim), dim, n_list,
                                       faiss.METRIC_INNER_PRODUCT)
            index.nprobe = min(math.ceil(index.nlist / 8), n_probe)
        # Compute cluster centroids.
        # noinspection PyArgumentList
        index.train(vectors_split)
        # Add the vectors to the index in batches.
        batch_size = min(n_split, batch_size)
        for batch_start in range(0, n_split, batch_size):
            batch_stop = min(batch_start + batch_size, n_split)
            # noinspection PyArgumentList
            index.add_with_ids(vectors_split[batch_start:batch_stop],
                               np.arange(batch_start, batch_stop))
        # Query the index to calculate NN distances.
        _dist_mz_interval(
            index, vectors_split, precursor_mzs[-1], rts[-1], batch_size,
            n_neighbors, n_neighbors_ann, precursor_tol_mass,
            precursor_tol_mode, rt_tol, distances, indices, indptr, indptr_i)
        index.reset()
        indptr_i += n_split
    if len(identifiers) == 0:
        return pd.DataFrame(columns=['identifier', 'precursor_mz',
                                     'retention_time'])
    else:
        return pd.DataFrame({'identifier': identifiers,
                             'precursor_mz': np.hstack(precursor_mzs),
                             'retention_time': np.hstack(rts)})


def _dist_mz_interval(
        index: faiss.Index, vectors: np.ndarray, precursor_mzs: np.ndarray,
        rts: np.ndarray, batch_size: int, n_neighbors: int,
        n_neighbors_ann: int, precursor_tol_mass: float,
        precursor_tol_mode: str, rt_tol: float, distances: np.ndarray,
        indices: np.ndarray, indptr: np.ndarray, indptr_i: int) -> None:
    """
    Compute distances to the nearest neighbors for the given precursor m/z
    interval.

    Parameters
    ----------
    index : faiss.Index
        The NN index used to efficiently find distances to similar spectra.
    vectors : np.ndarray
        The spectrum vectors to be queried against the NN index.
    precursor_mzs : np.ndarray
        Precorsor m/z's of the spectra corresponding to the given vectors.
    rts : np.ndarray
        Retention times corresponding to the vectors.
    batch_size : int
        The number of vectors to be simultaneously queried.
    n_neighbors : int
        The final (maximum) number of neighbors to retrieve for each vector.
    n_neighbors_ann : int
        The number of neighbors to retrieve using the ANN index. This can
        exceed the final number of neighbors (`n_neighbors`) to maximize the
        number of neighbors within the precursor m/z tolerance.
    precursor_tol_mass : float
        The precursor tolerance mass for vectors to be considered as neighbors.
    precursor_tol_mode : str
        The unit of the precursor m/z tolerance ('Da' or 'ppm').
    rt_tol : float
        The retention time tolerance for vectors to be considered as neighbors.
        If `None`, do not filter neighbors on retention time.
    distances : np.ndarray
        The nearest neighbor distances. See `scipy.sparse.csr_matrix` (`data`).
    indices : np.ndarray
        The column indices for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    indptr : np.ndarray
        The index pointers for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    indptr_i : int
        The current start index in `indptr`.
    """
    for batch_start in range(0, vectors.shape[0], batch_size):
        batch_stop = min(batch_start + batch_size, vectors.shape[0])
        # Find nearest neighbors using ANN index searching.
        # noinspection PyArgumentList
        nn_dists, nn_idx_ann = index.search(vectors[batch_start:batch_stop],
                                            n_neighbors_ann)
        # Filter the neighbors based on the precursor m/z tolerance and assign
        # distances.
        _filter_neighbors_mz(
            precursor_mzs, rts, batch_start, batch_stop, precursor_tol_mass,
            precursor_tol_mode, rt_tol, nn_dists, nn_idx_ann, n_neighbors,
            distances, indices, indptr, indptr_i + batch_start)


@nb.njit
def _filter_neighbors_mz(
        precursor_mzs: np.ndarray, rts: np.ndarray, batch_start: int,
        batch_stop: int, precursor_tol_mass: float, precursor_tol_mode: str,
        rt_tol: float, nn_dists: np.ndarray, nn_idx_ann: np.ndarray,
        n_neighbors: int, distances: np.ndarray, indices: np.ndarray,
        indptr: np.ndarray, indptr_i: int) -> None:
    """
    Filter ANN neighbor indexes by precursor m/z tolerances and assign
    pairwise distances.

    Parameters
    ----------
    precursor_mzs : np.ndarray
        Precursor m/z's corresponding to the vectors.
    rts : np.ndarray
        Retention times corresponding to the vectors.
    batch_start, batch_stop : int
        The indexes in the precursor m/z's of the current batch.
    precursor_tol_mass : float
        The precursor tolerance mass for vectors to be considered as neighbors.
    precursor_tol_mode : str
        The unit of the precursor m/z tolerance ('Da' or 'ppm').
    rt_tol : float
        The retention time tolerance for vectors to be considered as neighbors.
        If `None`, do not filter neighbors on retention time.
    nn_dists : np.ndarray
        Distances of the nearest neighbors.
    nn_idx_ann : np.ndarray
        Indexes of the nearest neighbors.
    n_neighbors : int
        The (maximum) number of neighbors to set for each vector.
    distances : np.ndarray
        The nearest neighbor distances. See `scipy.sparse.csr_matrix` (`data`).
    indices : np.ndarray
        The column indices for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    indptr : np.ndarray
        The index pointers for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    indptr_i : int
        The current start index in `indptr`.
    """
    nn_idx_mz = _get_neighbors_idx(
        precursor_mzs, batch_start, batch_stop,
        precursor_tol_mass, precursor_tol_mode)
    if rt_tol is not None:
        nn_idx_rt = _get_neighbors_idx(rts, batch_start, batch_stop, rt_tol,
                                       'rt')
        nn_idx_mz = [_intersect_idx_ann_mz(idx_mz, idx_rt, None, True)
                     for idx_mz, idx_rt in zip(nn_idx_mz, nn_idx_rt)]
    indptr_i_start = indptr_i - batch_start
    for i, (idx_ann, idx_mz, dists) in enumerate(
            zip(nn_idx_ann, nn_idx_mz, nn_dists), indptr_i):
        mask = _intersect_idx_ann_mz(idx_ann, idx_mz, n_neighbors)
        indptr[i + 1] = indptr[i] + len(mask)
        # Convert cosine similarity to cosine distance.
        distances[indptr[i]:indptr[i + 1]] = np.maximum(1 - dists[mask], 0)
        indices[indptr[i]:indptr[i + 1]] = indptr_i_start + idx_ann[mask]


@nb.njit
def _get_neighbors_idx(values: np.ndarray, start_i: int, stop_i: int,
                       tol: float, tol_mode: str) \
        -> List[np.ndarray]:
    """
    Filter nearest neighbor candidates on precursor m/z or retention time.

    Parameters
    ----------
    values : np.ndarray
        The precursor m/z (sorted) or retention time (unsorted) values of the
        candidates.
    start_i, stop_i : int
        Indexes used to slice the values to be considered in the batch
        (inclusive start_i, exclusive stop_i).
    tol : float
        The tolerance for vectors to be considered as neighbors.
    tol_mode : str
        The unit of the tolerance ('Da' or 'ppm' for precursor m/z, 'rt' for
        retention time).

    Returns
    -------
    List[np.ndarray]
        A list of sorted NumPy arrays with the indexes of the nearest neighbor
        candidates for each item.
    """
    batch_values = values[start_i:stop_i]
    if tol_mode in ('Da', 'ppm'):
        if tol_mode == 'Da':
            min_value = batch_values[0] - tol
            max_value = batch_values[-1] + tol
        elif tol_mode == 'ppm':
            min_value = batch_values[0] - batch_values[0] * tol / 10 ** 6
            max_value = batch_values[-1] + batch_values[-1] * tol / 10 ** 6
        # noinspection PyUnboundLocalVariable
        min_value, max_value = max(0, min_value), max(0, max_value)
        match_i = np.searchsorted(values, [min_value, max_value])
        match_values_i = np.arange(match_i[0], match_i[1])
        match_values = values[match_i[0]:match_i[1]].reshape((1, -1))
    elif tol_mode == 'rt':
        # RT values are not sorted.
        min_value = max(0, batch_values.min() - tol)
        max_value = max(0, batch_values.max() + tol)
        match_values_i = np.where(np.logical_and(values >= min_value,
                                                 values <= max_value))[0]
        match_values = values[match_values_i].reshape((1, -1))
    else:
        raise ValueError('Unknown tolerance filter')
    batch_values = batch_values.reshape((stop_i - start_i, 1))
    if tol_mode in ('Da', 'rt'):
        masks = np.abs(batch_values - match_values) < tol
    elif tol_mode == 'ppm':
        masks = (np.abs(batch_values - match_values) / match_values * 10 ** 6
                 < tol)
    # noinspection PyUnboundLocalVariable
    return [np.sort(match_values_i[mask]) for mask in masks]


@nb.njit
def _intersect_idx_ann_mz(idx_ann: np.ndarray, idx_mz: np.ndarray,
                          max_neighbors: int = None, is_sorted: bool = False) \
        -> np.ndarray:
    """
    Find the intersection between identifiers from ANN filtering and precursor
    m/z filtering.

    Parameters
    ----------
    idx_ann : np.ndarray
        Identifiers from ANN filtering.
    idx_mz : np.ndarray
        SORTED identifiers from precursor m/z filtering.
    max_neighbors : int
        The maximum number of best matching neighbors to retain.
    is_sorted : bool
        Flag indicating whether the first array is sorted or not.

    Returns
    -------
    np.ndarray
        A mask to select the joint identifiers in the `idx_ann` array.
    """
    i_mz = 0
    idx_ann_order = (np.argsort(idx_ann) if not is_sorted
                     else np.arange(len(idx_ann)))
    idx = []
    for i_order, i_ann in enumerate(idx_ann_order):
        if idx_ann[i_ann] != -1:
            while i_mz < len(idx_mz) and idx_mz[i_mz] < idx_ann[i_ann]:
                i_mz += 1
            if i_mz == len(idx_mz):
                break
            if idx_ann[i_ann] == idx_mz[i_mz]:
                idx.append(idx_ann_order[i_order])
                i_mz += 1
    idx = np.asarray(idx)
    if max_neighbors is None or max_neighbors >= len(idx):
        return idx
    else:
        return np.partition(idx, max_neighbors)[:max_neighbors]


def generate_clusters(pairwise_dist_matrix: ss.csr_matrix, eps: float,
                      min_samples: int, precursor_mzs: np.ndarray,
                      rts: np.ndarray, precursor_tol_mass: float,
                      precursor_tol_mode: str, rt_tol: float) \
        -> np.ndarray:
    """
    DBSCAN clustering of the given pairwise distance matrix.

    Parameters
    ----------
    pairwise_dist_matrix : ss.csr_matrix
        A sparse pairwise distance matrix used for clustering.
    eps : float
        The maximum distance between two samples for one to be considered as in
        the neighborhood of the other.
    min_samples : int
        The number of samples in a neighborhood for a point to be considered as
        a core point. This includes the point itself.
    precursor_mzs : np.ndarray
        Precursor m/z's matching the pairwise distance matrix.
    rts : np.ndarray
        Retention times matching the pairwise distance matrix.
    precursor_tol_mass : float
        Maximum precursor mass tolerance for points to be clustered together.
    precursor_tol_mode : str
        The unit of the precursor m/z tolerance ('Da' or 'ppm').
    rt_tol : float
        The retention time tolerance for points to be clustered together. If
        `None`, do not restrict the retention time.

    Returns
    -------
    np.ndarray
        Cluster labels. Noisy samples are given the label -1.
    """
    # DBSCAN clustering using the precomputed pairwise distance matrix.
    logger.debug('DBSCAN clustering (eps=%.4f, min_samples=%d) of precomputed '
                 'pairwise distance matrix', eps, min_samples)
    with tempfile.NamedTemporaryFile() as mmap_file:
        # Reimplement DBSCAN preprocessing to avoid unnecessary memory
        # consumption.
        # Find the eps-neighborhoods for all points.
        mask = pairwise_dist_matrix.data <= eps
        indices = pairwise_dist_matrix.indices[mask].astype(np.intp)
        indptr = np.zeros(len(mask) + 1, dtype=np.int64)
        np.cumsum(mask, out=indptr[1:])
        indptr = indptr[pairwise_dist_matrix.indptr]
        neighborhoods = np.split(indices, indptr[1:-1])
        # Initially, all samples are noise.
        # (Memmap for shared memory multiprocessing.)
        clusters = np.memmap(mmap_file, np.intp, 'w+',
                             shape=pairwise_dist_matrix.shape[0])
        clusters.fill(-1)
        # A list of all core samples found.
        n_neighbors = np.fromiter(map(len, neighborhoods), np.uint32)
        core_samples = n_neighbors >= min_samples
        # Run Scikit-Learn DBSCAN.
        neighborhoods_arr = np.empty(len(neighborhoods), dtype=np.object)
        neighborhoods_arr[:] = neighborhoods
        dbscan_inner(core_samples, neighborhoods_arr, clusters)
        # Refine initial clusters to make sure spectra within a cluster don't
        # have an excessive precursor m/z difference.
        order = np.argsort(clusters)
        reverse_order = np.argsort(order)
        clusters[:] = clusters[order]
        precursor_mzs, rts = precursor_mzs[order], rts[order]
        logger.debug('Finetune %d initial unique clusters to not exceed %.2f '
                     '%s precursor m/z tolerance%s', clusters[-1] + 1,
                     precursor_tol_mass, precursor_tol_mode,
                     f' and {rt_tol} retention time tolerance'
                     if rt_tol is not None else '')
        if clusters[-1] == -1:     # Only noise samples.
            clusters.fill(-1)
        else:
            group_idx = nb.typed.List(_get_cluster_group_idx(clusters))
            n_clusters = nb.typed.List(joblib.Parallel(n_jobs=-1)(
                joblib.delayed(_postprocess_cluster)
                (clusters[start_i:stop_i], precursor_mzs[start_i:stop_i],
                 rts[start_i:stop_i], precursor_tol_mass, precursor_tol_mode,
                 rt_tol, min_samples)
                for start_i, stop_i in group_idx))
            _assign_unique_cluster_labels(
                clusters, group_idx, n_clusters, min_samples)
            clusters = clusters[reverse_order]
        logger.debug('%d unique clusters after precursor m/z finetuning',
                     np.amax(clusters) + 1)
        return np.asarray(clusters)


@nb.njit
def _get_cluster_group_idx(clusters: np.ndarray) -> Iterator[Tuple[int, int]]:
    """
    Get start and stop indexes for unique cluster labels.
    Parameters
    ----------
    clusters : np.ndarray
        The ordered cluster labels (noise points are -1).
    Returns
    -------
    Iterator[Tuple[int, int]]
        Tuples with the start index (inclusive) and end index (exclusive) of
        the unique cluster labels.
    """
    start_i = 0
    while clusters[start_i] == -1 and start_i < clusters.shape[0]:
        start_i += 1
    stop_i = start_i
    while stop_i < clusters.shape[0]:
        start_i, label = stop_i, clusters[stop_i]
        while stop_i < clusters.shape[0] and clusters[stop_i] == label:
            stop_i += 1
        yield start_i, stop_i


def _postprocess_cluster(cluster_labels: np.ndarray, cluster_mzs: np.ndarray,
                         cluster_rts: np.ndarray, precursor_tol_mass: float,
                         precursor_tol_mode: str, rt_tol: float,
                         min_samples: int) -> int:
    """
    Agglomerative clustering of the precursor m/z's within each initial
    cluster to avoid that spectra within a cluster have an excessive precursor
    m/z difference.
    Parameters
    ----------
    cluster_labels : np.ndarray
        Array in which to write the cluster labels.
    cluster_mzs : np.ndarray
        Precursor m/z's of the samples in a single initial cluster.
    cluster_rts : np.ndarray
        Retention times of the samples in a single initial cluster.
    precursor_tol_mass : float
        Maximum precursor mass tolerance for points to be clustered together.
    precursor_tol_mode : str
        The unit of the precursor m/z tolerance ('Da' or 'ppm').
    rt_tol : float
        The retention time tolerance for points to be clustered together. If
        `None`, do not restrict the retention time.
    min_samples : int
        The minimum number of samples in a cluster.
    Returns
    -------
    int
        The number of clusters after splitting on precursor m/z.
    """
    cluster_labels.fill(-1)
    # No splitting needed if there are too few items in cluster.
    # This seems to happen sometimes despite that DBSCAN requires a higher
    # `min_samples`.
    if cluster_labels.shape[0] < min_samples:
        n_clusters = 0
    else:
        cluster_mzs = cluster_mzs.reshape(-1, 1)
        # Pairwise differences in Dalton.
        pairwise_mz_diff = pairwise_distances(cluster_mzs)
        if precursor_tol_mode == 'ppm':
            pairwise_mz_diff = pairwise_mz_diff / cluster_mzs * 10**6
        # Group items within the cluster based on their precursor m/z.
        # Precursor m/z's within a single group can't exceed the specified
        # precursor m/z tolerance (`distance_threshold`).
        # Subtract 1 because fcluster starts with cluster label 1 instead of 0
        # (like scikit-learn does).
        cluster_assignments = fcluster(
            fastcluster.linkage(
                squareform(pairwise_mz_diff, checks=False), 'complete'),
            precursor_tol_mass, 'distance') - 1
        # Optionally restrict clusters by their retention time as well.
        if rt_tol is not None:
            pairwise_rt_diff = pairwise_distances(cluster_rts.reshape(-1, 1))
            cluster_assignments_rt = fcluster(
                fastcluster.linkage(
                    squareform(pairwise_rt_diff, checks=False), 'complete'),
                rt_tol, 'distance') - 1
            # Merge cluster assignments based on precursor m/z and RT.
            # First prime factorization is used to get unique combined cluster
            # labels, after which consecutive labels are obtained.
            cluster_assignments = np.unique(
                cluster_assignments * 2 + cluster_assignments_rt * 3,
                return_inverse=True)[1]
        n_clusters = cluster_assignments.max() + 1
        # Update cluster assignments.
        if n_clusters == 1:
            # Single homogeneous cluster.
            cluster_labels.fill(0)
        elif n_clusters == cluster_mzs.shape[0]:
            # Only singletons.
            n_clusters = 0
        else:
            unique, inverse, counts = np.unique(
                cluster_assignments, return_inverse=True, return_counts=True)
            non_noise_clusters = np.where(counts >= min_samples)[0]
            labels = -np.ones_like(unique)
            labels[non_noise_clusters] = np.unique(unique[non_noise_clusters],
                                                   return_inverse=True)[1]
            cluster_labels[:] = labels[inverse]
            n_clusters = len(non_noise_clusters)
    return n_clusters


@nb.njit
def _assign_unique_cluster_labels(cluster_labels: np.ndarray,
                                  group_idx: nb.typed.List,
                                  n_clusters: nb.typed.List,
                                  min_samples: int) -> None:
    """
    Make sure all cluster labels are unique after potential splitting of
    clusters to avoid excessive precursor m/z differences.

    Parameters
    ----------
    cluster_labels : np.ndarray
        Cluster labels per cluster grouping.
    group_idx : nb.typed.List[Tuple[int, int]]
        Tuples with the start index (inclusive) and end index (exclusive) of
        the cluster groupings.
    n_clusters: nb.typed.List[int]
        The number of clusters per cluster grouping.
    min_samples : int
        The minimum number of samples in a cluster.
    """
    current_label = 0
    for (start_i, stop_i), n_cluster in zip(group_idx, n_clusters):
        if n_cluster > 0 and stop_i - start_i >= min_samples:
            current_labels = cluster_labels[start_i:stop_i]
            current_labels[current_labels != -1] += current_label
            current_label += n_cluster
        else:
            cluster_labels[start_i:stop_i].fill(-1)


@nb.njit(parallel=True)
def get_cluster_representatives(clusters: np.ndarray,
                                pairwise_indptr: np.ndarray,
                                pairwise_indices: np.ndarray,
                                pairwise_data: np.ndarray) \
        -> Optional[np.ndarray]:
    """
    Get indexes of the cluster representative spectra (medoids).

    Parameters
    ----------
    clusters : np.ndarray
        Cluster label assignments.
    pairwise_indptr : np.ndarray
        The index pointers for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    pairwise_indices : np.ndarray
        The column indices for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    pairwise_data : np.ndarray
        The nearest neighbor distances. See `scipy.sparse.csr_matrix` (`data`).

    Returns
    -------
    Optional[np.ndarray]
        The indexes of the medoid elements for all non-noise clusters, or None
        if only noise clusters are present.
    """
    order, min_i = np.argsort(clusters), 0
    while min_i < clusters.shape[0] and clusters[order[min_i]] == -1:
        min_i += 1
    # Only noise clusters.
    if min_i == clusters.shape[0]:
        return None
    # Find the indexes of the representatives for each unique cluster.
    cluster_idx, max_i = [], min_i
    while max_i < order.shape[0]:
        while (max_i < order.shape[0] and
               clusters[order[min_i]] == clusters[order[max_i]]):
            max_i += 1
        cluster_idx.append((min_i, max_i))
        min_i = max_i
    representatives = np.empty(len(cluster_idx), np.uint)
    for i in nb.prange(len(cluster_idx)):
        representatives[i] = _get_cluster_medoid_index(
            order[cluster_idx[i][0]:cluster_idx[i][1]], pairwise_indptr,
            pairwise_indices, pairwise_data)
    return representatives


@nb.njit(fastmath=True)
def _get_cluster_medoid_index(cluster_mask: np.ndarray,
                              pairwise_indptr: np.ndarray,
                              pairwise_indices: np.ndarray,
                              pairwise_data: np.ndarray) -> int:
    """
    Get the index of the cluster medoid element.

    Parameters
    ----------
    cluster_mask : np.ndarray
        Indexes of the items belonging to the current cluster.
    pairwise_indptr : np.ndarray
        The index pointers for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    pairwise_indices : np.ndarray
        The column indices for the nearest neighbor distances. See
        `scipy.sparse.csr_matrix`.
    pairwise_data : np.ndarray
        The nearest neighbor distances. See `scipy.sparse.csr_matrix` (`data`).

    Returns
    -------
    int
        The index of the cluster's medoid element.
    """
    if len(cluster_mask) <= 2:
        # Pairwise distances will be identical.
        return cluster_mask[0]
    min_i, min_avg = 0, np.inf
    for row_i in range(cluster_mask.shape[0]):
        indices = pairwise_indices[pairwise_indptr[cluster_mask[row_i]]:
                                   pairwise_indptr[cluster_mask[row_i] + 1]]
        data = pairwise_data[pairwise_indptr[cluster_mask[row_i]]:
                             pairwise_indptr[cluster_mask[row_i] + 1]]
        col_i = np.asarray([i for cm in cluster_mask
                            for i, ind in enumerate(indices) if cm == ind])
        row_avg = np.mean(data[col_i]) if len(col_i) > 0 else np.inf
        if row_avg < min_avg:
            min_i, min_avg = row_i, row_avg
    return cluster_mask[min_i]



import logging
import os
from typing import Dict, IO, Iterable, Union

import pyteomics.mzxml
import spectrum_utils.spectrum as sus
from lxml.etree import LxmlError


logger = logging.getLogger('falcon')


def get_spectra(source: Union[IO, str]) -> Iterable[sus.MsmsSpectrum]:
    """
    Get the MS/MS spectra from the given mzXML file.

    Parameters
    ----------
    source : Union[IO, str]
        The mzXML source (file name or open file object) from which the spectra
        are read.

    Returns
    -------
    Iterable[MsmsSpectrum]
        An iterator over the spectra in the given file.
    """
    with pyteomics.mzxml.MzXML(source) as f_in:
        filename = os.path.splitext(os.path.basename(f_in.name))[0]
        try:
            for spectrum_dict in f_in:
                if int(spectrum_dict.get('msLevel', -1)) == 2:
                    # USI-inspired cluster identifier.
                    scan_nr = int(spectrum_dict['id'])
                    spectrum_dict['id'] = f'{filename}:scan:{scan_nr}'
                    try:
                        yield _parse_spectrum(spectrum_dict)
                    except (ValueError, KeyError):
                        pass
        except LxmlError as e:
            logger.warning('Failed to read file %s: %s', source, e)


def _parse_spectrum(spectrum_dict: Dict) -> sus.MsmsSpectrum:
    """
    Parse the Pyteomics cluster dict.

    Parameters
    ----------
    spectrum_dict : Dict
        The Pyteomics cluster dict to be parsed.

    Returns
    -------
    MsmsSpectrum
        The parsed cluster.
    """
    spectrum_id = spectrum_dict['id']
    mz_array = spectrum_dict['m/z array']
    intensity_array = spectrum_dict['intensity array']
    retention_time = spectrum_dict.get('retentionTime', -1)

    precursor_mz = spectrum_dict['precursorMz'][0]['precursorMz']
    if 'precursorCharge' in spectrum_dict['precursorMz'][0]:
        precursor_charge = spectrum_dict['precursorMz'][0]['precursorCharge']
    else:
        raise ValueError('Unknown precursor charge')

    return sus.MsmsSpectrum(spectrum_id, precursor_mz, precursor_charge,
                            mz_array, intensity_array, None, retention_time)



import os
from typing import Iterable, Iterator

from spectrum_utils.spectrum import MsmsSpectrum

from . import mzxml_io
from . import mgf_io
from . import mzml_io


def get_spectra(filename: str) -> Iterator[MsmsSpectrum]:
    """
    Get the MS/MS spectra from the given file.

    Supported file formats are MGF, MSP, mzML, mzXML.

    Parameters
    ----------
    filename : str
        The file name from which to read the spectra.

    Returns
    -------
    Iterator[MsmsSpectrum]
        An iterator over the spectra in the given file.
    """
    if not os.path.isfile(filename):
        raise ValueError(f'Non-existing peak file {filename}')

    _, ext = os.path.splitext(filename.lower())
    if ext == '.mgf':
        spectrum_io = mgf_io
    elif ext == '.mzml':
        spectrum_io = mzml_io
    elif ext == '.mzxml':
        spectrum_io = mzxml_io
    else:
        raise ValueError(f'Unknown spectrum file type with extension "{ext}"')

    for spec in spectrum_io.get_spectra(filename):
        spec.is_processed = False
        yield spec


def write_spectra(filename: str, spectra: Iterable[MsmsSpectrum]) -> None:
    """
    Write the given spectra to a peak file.

    Supported formats: MGF.

    Parameters
    ----------
    filename : str
        The file name where the spectra will be written.
    spectra : Iterable[MsmsSpectrum]
        The spectra to be written to the peak file.
    """
    ext = os.path.splitext(filename.lower())[1]
    if ext == '.mgf':
        spectrum_io = mgf_io
    else:
        raise ValueError('Unsupported peak file format (supported formats: '
                         'MGF)')

    spectrum_io.write_spectra(filename, spectra)



import collections
import math
from typing import List, Optional, Tuple

import faiss
import joblib
import numba as nb
import numpy as np
import spectrum_utils.spectrum as sus


MsmsSpectrumNb = collections.namedtuple(
    'MsmsSpectrumNb', ['identifier', 'precursor_mz', 'precursor_charge',
                       'retention_time', 'mz', 'intensity'])


@nb.njit
def _check_spectrum_valid(spectrum_mz: np.ndarray, min_peaks: int,
                          min_mz_range: float) -> bool:
    """
    Check whether a cluster is of good enough quality to be used.

    Parameters
    ----------
    spectrum_mz : np.ndarray
        M/z peaks of the cluster whose quality is checked.
    min_peaks : int
        Minimum number of peaks the cluster has to contain.
    min_mz_range : float
        Minimum m/z range the cluster's peaks need to cover.

    Returns
    -------
    bool
        True if the cluster has enough peaks covering a wide enough mass
        range, False otherwise.
    """
    return (len(spectrum_mz) >= min_peaks and
            spectrum_mz[-1] - spectrum_mz[0] >= min_mz_range)


@nb.njit
def _norm_intensity(spectrum_intensity: np.ndarray) -> np.ndarray:
    """
    Normalize cluster peak intensities by their vector norm.

    Parameters
    ----------
    spectrum_intensity : np.ndarray
        The cluster peak intensities to be normalized.

    Returns
    -------
    np.ndarray
        The normalized peak intensities.
    """
    return spectrum_intensity / np.linalg.norm(spectrum_intensity)


def process_spectrum(spectrum: sus.MsmsSpectrum,
                     min_peaks: int, min_mz_range: float,
                     mz_min: Optional[float] = None,
                     mz_max: Optional[float] = None,
                     remove_precursor_tolerance: Optional[float] = None,
                     min_intensity: Optional[float] = None,
                     max_peaks_used: Optional[int] = None,
                     scaling: Optional[str] = None) \
        -> Optional[MsmsSpectrumNb]:
    """
    Process a cluster.

    Processing steps include:
    - Restrict the m/z range to a minimum and maximum m/z.
    - Remove peak(s) around the precursor m/z value.
    - Remove peaks below a percentage of the base peak intensity.
    - Retain only the top most intense peaks.
    - Scale and normalize peak intensities.

    Parameters
    ----------
    spectrum : MsmsSpectrum
        The cluster to be processed.
    min_peaks : int
        Minimum number of peaks the cluster has to contain to be valid.
    min_mz_range : float
        Minimum m/z range the cluster's peaks need to cover to be valid.
    mz_min : Optional[float], optional
        Minimum m/z (inclusive). If not set no minimal m/z restriction will
        occur.
    mz_max : Optional[float], optional
        Maximum m/z (inclusive). If not set no maximal m/z restriction will
        occur.
    remove_precursor_tolerance : Optional[float], optional
        Fragment mass tolerance (in Dalton) around the precursor mass to remove
        the precursor peak. If not set, the precursor peak will not be removed.
    min_intensity : Optional[float], optional
        Remove peaks whose intensity is below `min_intensity` percentage
        of the base peak intensity. If None, no minimum intensity filter will
        be applied.
    max_peaks_used : Optional[int], optional
        Only retain the `max_peaks_used` most intense peaks. If None, all peaks
        are retained.
    scaling : {'root', 'log', 'rank'}, optional
        Method to scale the peak intensities. Potential transformation options
        are:

        - 'root': Square root-transform the peak intensities.
        - 'log':  Log2-transform (after summing the intensities with 1 to avoid
          negative values after the transformation) the peak intensities.
        - 'rank': Rank-transform the peak intensities with maximum rank
          `max_peaks_used`.
        - None: No scaling is performed.

    Returns
    -------
    MsmsSpectrumNb
        The processed cluster.
    """
    spectrum = spectrum.set_mz_range(mz_min, mz_max)
    if not _check_spectrum_valid(spectrum.mz, min_peaks, min_mz_range):
        return None

    if remove_precursor_tolerance is not None:
        spectrum = spectrum.remove_precursor_peak(
            remove_precursor_tolerance, 'Da', 0)
        if not _check_spectrum_valid(spectrum.mz, min_peaks, min_mz_range):
            return None

    if min_intensity is not None or max_peaks_used is not None:
        min_intensity = 0. if min_intensity is None else min_intensity
        spectrum = spectrum.filter_intensity(min_intensity, max_peaks_used)
        if not _check_spectrum_valid(spectrum.mz, min_peaks, min_mz_range):
            return None

    spectrum = spectrum.scale_intensity(scaling, max_rank=max_peaks_used)
    intensity = _norm_intensity(spectrum.intensity)

    return MsmsSpectrumNb(spectrum.identifier, spectrum.precursor_mz,
                          spectrum.precursor_charge, spectrum.retention_time,
                          spectrum.mz, intensity)


@nb.njit('Tuple((u4, f4, f4))(f4, f4, f4)')
def get_dim(min_mz: float, max_mz: float, bin_size: float) \
        -> Tuple[int, float, float]:
    """
    Compute the number of bins over the given mass range for the given bin
    size.

    Parameters
    ----------
    min_mz : float
        The minimum mass in the mass range (inclusive).
    max_mz : float
        The maximum mass in the mass range (inclusive).
    bin_size : float
        The bin size (in Da).

    Returns
    -------
        A tuple containing (i) the number of bins over the given mass range for
        the given bin size, (ii) the highest multiple of bin size lower than
        the minimum mass, (iii) the lowest multiple of the bin size greater
        than the maximum mass. These two final values are the true boundaries
        of the mass range (inclusive min, exclusive max).
    """
    start_dim = min_mz - min_mz % bin_size
    end_dim = max_mz + bin_size - max_mz % bin_size
    return math.ceil((end_dim - start_dim) / bin_size), start_dim, end_dim


@nb.njit
def to_vector(spectrum: MsmsSpectrumNb, vector: np.ndarray,
              min_mz: float, max_mz: float, bin_size: float,
              hash_lookup: np.ndarray, norm: bool = True) -> np.ndarray:
    """
    Convert a spectrum to a hashed NumPy vector.

    Peaks are first discretized to mass bins of width `bin_size` between
    `min_mz` and `max_mz`, after which they are hashed to random hash bins
    (using the hash lookup table) in the final vector.

    Parameters
    ----------
    spectrum : MsmsSpectrumNb
        The spectrum to be converted to a vector.
    vector : np.ndarray
        A pre-allocated vector to store the output.
    min_mz : float
        The minimum m/z to include in the vector.
    max_mz : float
        The maximum m/z to include in the vector.
    bin_size : float
        The bin size in m/z used to divide the m/z range.
    hash_lookup : np.ndarray
        A lookup vector with hash indexes.
    norm : bool
        Normalize the vector to unit length or not.

    Returns
    -------
    np.ndarray
        The hashed spectrum vector.
    """
    _, min_bound, max_bound = get_dim(min_mz, max_mz, bin_size)
    for mz, intensity in zip(spectrum.mz, spectrum.intensity):
        if min_bound < mz < max_bound:
            hash_idx = hash_lookup[math.floor((mz - min_bound) / bin_size)]
            vector[hash_idx] += intensity
    return vector if not norm else _norm_intensity(vector)


def to_vector_parallel(spectra: List[MsmsSpectrumNb], dim: int, min_mz: float,
                       max_mz: float, bin_size: float, hash_lookup: np.ndarray,
                       norm: bool) -> np.ndarray:
    """
    Convert multiple spectra to hashed NumPy vectors.

    Peaks are first discretized to mass bins of width `bin_size` between
    `min_mz` and `max_mz`, after which they are hashed to random hash bins
    (using the hash lookup table) in the final vector.

    Parameters
    ----------
    spectra : List[MsmsSpectrumNb]
        The spectra to be converted to vectors.
    dim: int
        The dimensionality of the hashed spectrum vectors.
    min_mz : float
        The minimum m/z to include in the vector.
    max_mz : float
        The maximum m/z to include in the vector.
    bin_size : float
        The bin size in m/z used to divide the m/z range.
    hash_lookup : np.ndarray
        A lookup vector with hash indexes.
    norm : bool
        Normalize the vector to unit length or not.

    Returns
    -------
    np.ndarray
        The hashed spectrum vectors.
    """
    vectors = np.zeros((len(spectra), dim), np.float32)
    joblib.Parallel(n_jobs=-1, prefer='threads')(
        joblib.delayed(to_vector)(spec, vectors[i, :], min_mz, max_mz,
                                  bin_size, hash_lookup, False)
        for i, spec in enumerate(spectra))
    if norm:
        # Normalize the vectors for inner product search.
        faiss.normalize_L2(vectors)
    return vectors



import collections
import functools
import glob
import logging
import os
import pickle
import shutil
import sys
import tempfile
from typing import Dict, List, Tuple, Union

import joblib
import natsort
import numba as nb
import numpy as np
import pandas as pd
import scipy.sparse as ss
from sklearn.utils import murmurhash3_32
from spectrum_utils.spectrum import MsmsSpectrum

from . import __version__
from .cluster import cluster, spectrum
from .config import config
from .ms_io import ms_io


logger = logging.getLogger('falcon')


def main(args: Union[str, List[str]] = None) -> int:
    # Configure logging.
    logging.captureWarnings(True)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(
        '{asctime} {levelname} [{name}/{processName}] {module}.{funcName} : '
        '{message}', style='{'))
    root.addHandler(handler)
    # Disable dependency non-critical log messages.
    logging.getLogger('faiss').setLevel(logging.WARNING)
    logging.getLogger('numba').setLevel(logging.WARNING)
    logging.getLogger('numexpr').setLevel(logging.WARNING)

    # Load the configuration.
    config.parse(args)
    logger.info('falcon version %s', str(__version__))
    logger.debug('work_dir = %s', config.work_dir)
    logger.debug('overwrite = %s', config.overwrite)
    logger.debug('export_representatives = %s', config.export_representatives)
    logger.debug('usi_pxd = %s', config.usi_pxd)
    logger.debug('precursor_tol = %.2f %s', *config.precursor_tol)
    logger.debug('rt_tol = %s', config.rt_tol)
    logger.debug('fragment_tol = %.2f', config.fragment_tol)
    logger.debug('eps = %.3f', config.eps)
    logger.debug('min_samples = %d', config.min_samples)
    logger.debug('mz_interval = %d', config.mz_interval)
    logger.debug('hash_len = %d', config.hash_len)
    logger.debug('n_neighbors = %d', config.n_neighbors)
    logger.debug('n_neighbors_ann = %d', config.n_neighbors_ann)
    logger.debug('batch_size = %d', config.batch_size)
    logger.debug('n_probe = %d', config.n_probe)
    logger.debug('min_peaks = %d', config.min_peaks)
    logger.debug('min_mz_range = %.2f', config.min_mz_range)
    logger.debug('min_mz = %.2f', config.min_mz)
    logger.debug('max_mz = %.2f', config.max_mz)
    logger.debug('remove_precursor_tol = %.2f', config.remove_precursor_tol)
    logger.debug('min_intensity = %.2f', config.min_intensity)
    logger.debug('max_peaks_used = %d', config.max_peaks_used)
    logger.debug('scaling = %s', config.scaling)

    rm_work_dir = False
    if config.work_dir is None:
        config.work_dir = tempfile.mkdtemp()
        rm_work_dir = True
    elif os.path.isdir(config.work_dir):
        logging.warning('Working directory %s already exists, previous '
                        'results might get overwritten', config.work_dir)
    os.makedirs(config.work_dir, exist_ok=True)
    os.makedirs(os.path.join(config.work_dir, 'spectra'), exist_ok=True)
    os.makedirs(os.path.join(config.work_dir, 'nn'), exist_ok=True)

    # Clean all intermediate and final results if "overwrite" is specified,
    # otherwise abort if the output files already exist.
    exit_exists = False
    if os.path.isfile(f'{config.output_filename}.csv'):
        if config.overwrite:
            logger.warning('Output file %s (cluster assignments) already '
                           'exists and will be overwritten',
                           f'{config.output_filename}.csv')
            os.remove(f'{config.output_filename}.csv')
        else:
            logger.error('Output file %s (cluster assignments) already '
                         'exists, aborting...',
                         f'{config.output_filename}.csv')
            exit_exists = True
    if os.path.isfile(f'{config.output_filename}.mgf'):
        if config.overwrite:
            logger.warning('Output file %s (cluster representatives) already '
                           'exists and will be overwritten',
                           f'{config.output_filename}.mgf')
            os.remove(f'{config.output_filename}.mgf')
        else:
            logger.error('Output file %s (cluster representatives) already '
                         'exists, aborting...',
                         f'{config.output_filename}.mgf')
            exit_exists = True
    if exit_exists:
        logging.shutdown()
        return 1
    if config.overwrite:
        for filename in os.listdir(os.path.join(config.work_dir, 'spectra')):
            os.remove(os.path.join(config.work_dir, 'spectra', filename))
        for filename in os.listdir(os.path.join(config.work_dir, 'nn')):
            os.remove(os.path.join(config.work_dir, 'nn', filename))

    # Read the spectra from the input files and partition them based on their
    # precursor m/z.
    if not any([filename.endswith('.pkl') for filename in os.listdir(
            os.path.join(config.work_dir, 'spectra'))]):
        buckets = _prepare_spectra()
        joblib.dump(buckets, os.path.join(config.work_dir, 'spectra',
                                          'buckets.joblib'))
    else:
        buckets = joblib.load(os.path.join(config.work_dir, 'spectra',
                                           'buckets.joblib'))

    vec_len, min_mz, max_mz = spectrum.get_dim(config.min_mz, config.max_mz,
                                               config.fragment_tol)
    process_spectrum = functools.partial(
        spectrum.process_spectrum,
        min_peaks=config.min_peaks,
        min_mz_range=config.min_mz_range,
        mz_min=min_mz,
        mz_max=max_mz,
        remove_precursor_tolerance=config.remove_precursor_tol,
        min_intensity=config.min_intensity,
        max_peaks_used=config.max_peaks_used,
        scaling=None if config.scaling == 'off' else config.scaling)

    # Pre-compute the index hash mappings.
    hash_lookup = np.asarray([murmurhash3_32(i, 0, True) % config.hash_len
                              for i in range(vec_len)], np.uint32)
    vectorize = functools.partial(
        spectrum.to_vector_parallel, dim=config.hash_len, min_mz=min_mz,
        max_mz=max_mz, bin_size=config.fragment_tol, hash_lookup=hash_lookup,
        norm=True)

    # Cluster the spectra per charge.
    clusters_all, current_label, representative_info = [], 0, []
    for charge, (n_spectra, bucket_filenames) in buckets.items():
        logger.info('Cluster %d spectra with precursor charge %d',
                    n_spectra, charge)
        dist_filename = os.path.join(
            config.work_dir, 'nn', f'dist_{charge}.npz')
        metadata_filename = os.path.join(
            config.work_dir, 'nn', f'metadata_{charge}.parquet')
        if (not os.path.isfile(dist_filename)
                or not os.path.isfile(metadata_filename)):
            pairwise_dist_matrix, metadata = \
                cluster.compute_pairwise_distances(
                    n_spectra, bucket_filenames, process_spectrum, vectorize,
                    config.precursor_tol[0], config.precursor_tol[1],
                    config.rt_tol, config.n_neighbors, config.n_neighbors_ann,
                    config.batch_size, config.n_probe)
            metadata.insert(1, 'precursor_charge', charge)
            logger.debug('Export pairwise distance matrix to file %s',
                         dist_filename)
            ss.save_npz(dist_filename, pairwise_dist_matrix, False)
            metadata.to_parquet(metadata_filename, index=False)
        else:
            logger.debug('Load previously computed pairwise distance matrix '
                         'from file %s', dist_filename)
            pairwise_dist_matrix = ss.load_npz(dist_filename)
            metadata = pd.read_parquet(metadata_filename)
        # No valid spectra found with the current charge.
        if len(metadata) == 0:
            continue
        # Cluster using the pairwise distance matrix.
        clusters = cluster.generate_clusters(
            pairwise_dist_matrix, config.eps, config.min_samples,
            metadata['precursor_mz'].values, metadata['retention_time'].values,
            config.precursor_tol[0], config.precursor_tol[1], config.rt_tol)
        # Make sure that different charges have non-overlapping cluster labels.
        mask_no_noise = clusters != -1
        clusters[mask_no_noise] += current_label
        current_label = (np.amax(clusters[mask_no_noise]) + 1
                         if any(mask_no_noise) else current_label)
        # Save cluster assignments.
        metadata['cluster'] = clusters
        clusters_all.append(metadata)
        # Extract identifiers for cluster representatives (medoids).
        if config.export_representatives:
            charge_repr = cluster.get_cluster_representatives(
                clusters, pairwise_dist_matrix.indptr,
                pairwise_dist_matrix.indices, pairwise_dist_matrix.data)
            if charge_repr is not None:
                representative_info.append(metadata.iloc[charge_repr])
            if config.export_include_singletons:
                representative_info.append(metadata.loc[~mask_no_noise])
            logger.debug('Extract %d cluster representative %sidentifiers',
                         len(charge_repr) if charge_repr is not None else 0,
                         f'and {(~mask_no_noise).sum()} singleton spectra '
                         if config.export_include_singletons else '')

    # Export cluster memberships and representative spectra.
    n_clusters, n_spectra_clustered = 0, 0
    for clust in clusters_all:
        clust_no_noise = clust[clust['cluster'] != -1]
        n_clusters += clust_no_noise['cluster'].nunique()
        n_spectra_clustered += len(clust_no_noise)
    if n_spectra_clustered == 0:
        logger.error('No valid spectra found for clustering')
        logging.shutdown()
        return 1
    logger.info('Export cluster assignments of %d spectra to %d unique '
                'clusters to output file %s', n_spectra_clustered, n_clusters,
                f'{config.output_filename}.csv')
    clusters_all = (pd.concat(clusters_all, ignore_index=True)
                    .sort_values('identifier', key=natsort.natsort_keygen()))
    with open(f'{config.output_filename}.csv', 'a') as f_out:
        # Metadata.
        f_out.write(f'# falcon version {__version__}\n')
        f_out.write(f'# work_dir = {config.work_dir}\n')
        f_out.write(f'# overwrite = {config.overwrite}\n')
        f_out.write(f'# export_representatives = '
                    f'{config.export_representatives}\n')
        f_out.write(f'# usi_pxd = {config.usi_pxd}\n')
        f_out.write(f'# precursor_tol = {config.precursor_tol[0]:.2f} '
                    f'{config.precursor_tol[1]}\n')
        f_out.write(f'# rt_tol = {config.rt_tol}\n')
        f_out.write(f'# fragment_tol = {config.fragment_tol:.2f}\n')
        f_out.write(f'# eps = {config.eps:.3f}\n')
        f_out.write(f'# min_samples = {config.min_samples}\n')
        f_out.write(f'# mz_interval = {config.mz_interval}\n')
        f_out.write(f'# hash_len = {config.hash_len}\n')
        f_out.write(f'# n_neighbors = {config.n_neighbors}\n')
        f_out.write(f'# n_neighbors_ann = {config.n_neighbors_ann}\n')
        f_out.write(f'# batch_size = {config.batch_size}\n')
        f_out.write(f'# n_probe = {config.n_probe}\n')
        f_out.write(f'# min_peaks = {config.min_peaks}\n')
        f_out.write(f'# min_mz_range = {config.min_mz_range:.2f}\n')
        f_out.write(f'# min_mz = {config.min_mz:.2f}\n')
        f_out.write(f'# max_mz = {config.max_mz:.2f}\n')
        f_out.write(f'# remove_precursor_tol = '
                    f'{config.remove_precursor_tol:.2f}\n')
        f_out.write(f'# min_intensity = {config.min_intensity:.2f}\n')
        f_out.write(f'# max_peaks_used = {config.max_peaks_used}\n')
        f_out.write(f'# scaling = {config.scaling}\n')
        f_out.write('#\n')
        # Cluster assignments.
        clusters_all.to_csv(f_out, index=False, chunksize=1000000)
    if config.export_representatives:
        representative_info = pd.concat(representative_info, ignore_index=True)
        logger.info('Export %d cluster representative spectra %sto output '
                    'file %s', len(representative_info),
                    ('(including singletons) '
                     if config.export_include_singletons else ''),
                    f'{config.output_filename}.mgf')
        # Get the spectra corresponding to the cluster representatives.
        representative_info['filename'] = representative_info.apply(
            lambda row: os.path.join(
                config.work_dir, 'spectra',
                f"""{row.precursor_charge}_{_precursor_to_interval(
                    row.precursor_mz, row.precursor_charge,
                    config.mz_interval)}.pkl"""),
            axis='columns')
        representatives = []
        for spectra in joblib.Parallel(n_jobs=-1)(
                joblib.delayed(_find_spectra_pkl)(
                    fn, spectra.set_index('identifier')['cluster'].to_dict())
                for fn, spectra in representative_info.groupby('filename')):
            representatives.extend(spectra)
        representatives.sort(key=lambda spec: spec.cluster)
        ms_io.write_spectra(f'{config.output_filename}.mgf', representatives)

    if rm_work_dir:
        shutil.rmtree(config.work_dir)

    logging.shutdown()
    return 0


def _prepare_spectra() -> Dict[int, Tuple[int, List[str]]]:
    """
    Read the spectra from the input peak files and partition to intermediate
    files split and sorted by precursor m/z.

    Returns
    -------
    Dict[int, Tuple[int, List[str]]]
        A dictionary with per precursor charge the total number of spectra and
        a list of file names of the spectrum buckets.
    """
    input_filenames = [fn for pattern in config.input_filenames
                       for fn in glob.glob(pattern)]
    logger.info('Read spectra from %d peak file(s)', len(input_filenames))
    spectra = collections.defaultdict(lambda: collections.defaultdict(list))
    for file_spectra in joblib.Parallel(n_jobs=-1)(
            joblib.delayed(_read_spectra)(filename, config.usi_pxd,
                                          config.mz_interval, config.work_dir)
            for filename in input_filenames):
        for spec, pkl_filename in file_spectra:
            spectra[spec.precursor_charge][pkl_filename].append(spec)
    # Make sure the spectra in the individual files are sorted by their
    # precursor m/z and count the number of spectra per precursor charge.
    buckets, n_spectra, n_buckets = {}, 0, 0
    for charge in sorted(spectra.keys()):
        spectra_charge = spectra[charge]
        n_spectra_charge = sum(
            joblib.Parallel(n_jobs=-1)
            (joblib.delayed(_write_spectra_pkl)(filename, file_spectra)
             for filename, file_spectra in spectra_charge.items()))
        buckets[charge] = (n_spectra_charge,
                           natsort.natsorted(spectra_charge.keys()))
        n_spectra += n_spectra_charge
        n_buckets += len(spectra_charge.keys())
    logger.debug('%d spectra written to %d buckets by precursor charge and '
                 'precursor m/z', n_spectra, n_buckets)
    return buckets


def _read_spectra(filename: str, usi_pxd: str, mz_interval: int,
                  work_dir: str) -> List[Tuple[MsmsSpectrum, str]]:
    """
    Get MS/MS spectra from the given file.

    Parameters
    ----------
    filename : str
        The path of the peak file to be read.
    usi_pxd : str
        ProteomeXchange dataset identifier to create USI references.
    mz_interval : int
        The width of each m/z interval.
    work_dir : str
        The directory in which the spectrum buckets will be stored.


    Returns
    -------
    List[Tuple[MsmsSpectrum, str]]
        The spectra read from the given file and their bucket filenames
        (based on precursor charge and m/z).
    """
    spectra = []
    for spec in ms_io.get_spectra(filename):
        spec.identifier = f'mzspec:{usi_pxd}:{spec.identifier}'
        interval = _precursor_to_interval(
            spec.precursor_mz, spec.precursor_charge, mz_interval)
        pkl_filename = os.path.join(work_dir, 'spectra',
                                    f'{spec.precursor_charge}_{interval}.pkl')
        spectra.append((spec, pkl_filename))
    return spectra


@nb.njit
def _precursor_to_interval(mz: float, charge: int, interval_width: int) -> int:
    """
    Convert the precursor m/z to the neutral mass and get the interval index.

    Parameters
    ----------
    mz : float
        The precursor m/z.
    charge : int
        The precursor charge.
    interval_width : int
        The width of each m/z interval.

    Returns
    -------
    int
        The index of the interval to which a spectrum with the given m/z and
        charge belongs.
    """
    hydrogen_mass, cluster_width = 1.00794, 1.0005079
    neutral_mass = (mz - hydrogen_mass) * charge
    return round(neutral_mass / cluster_width) // interval_width


def _write_spectra_pkl(filename: str, spectra: List[MsmsSpectrum]) -> int:
    """
    Write the spectra to a pickled file sorted by their precursor m/z.

    Parameters
    ----------
    filename : str
        The pickled spectrum file name.
    spectra : List[MsmsSpectrum]
        The spectra to write to the file.

    Returns
    -------
    int
        The number of spectra in the file.
    """
    # Radix sort on integers with O(n) time complexity.
    order = np.argsort(np.asarray([int(spec.precursor_mz * 10000)
                                   for spec in spectra], dtype=np.int32),
                       kind='stable')
    with open(filename, 'wb') as f_in:
        pickle.dump([spectra[i] for i in order], f_in, protocol=5)
    return len(spectra)


def _find_spectra_pkl(filename: str, usis_to_read: Dict[str, int]) \
        -> List[MsmsSpectrum]:
    """
    Read spectra with the specified USIs from a pkl file.

    Parameters
    ----------
    filename : str
        Name of the pkl file from which the spectra are read.
    usis_to_read : Dict[str, int]
        Dict with as keys the USIs of the spectra to read from the pkl file,
        and as values the cluster labels that will be assigned to the spectra.

    Returns
    -------
    Iterable[MsmsSpectrum]
        The spectra with the specified USIs.
    """
    spectra_found = []
    with open(filename, 'rb') as f_in:
        for spec in pickle.load(f_in):
            if spec.identifier in usis_to_read:
                spec.cluster = usis_to_read[spec.identifier]
                spectra_found.append(spec)
                if len(spectra_found) == len(usis_to_read):
                    return spectra_found
    raise ValueError(f'{len(usis_to_read) - len(spectra_found)} spectra not '
                     f'found in file {filename}')


if __name__ == '__main__':
    sys.exit(main())



import os
from typing import Dict, IO, Iterable, Union

import pyteomics.mgf
import spectrum_utils.spectrum as sus


def get_spectra(source: Union[IO, str]) -> Iterable[sus.MsmsSpectrum]:
    """
    Get the MS/MS spectra from the given MGF file.

    Parameters
    ----------
    source : Union[IO, str]
        The MGF source (file name or open file object) from which the spectra
        are read.

    Returns
    -------
    Iterator[MsmsSpectrum]
        An iterator over the spectra in the given file.
    """
    with pyteomics.mgf.MGF(source) as f_in:
        filename = os.path.splitext(os.path.basename(f_in.name))[0]
        for spectrum_i, spectrum_dict in enumerate(f_in):
            # USI-inspired cluster identifier.
            if 'scans' in spectrum_dict['params']:
                # Use a scan number as identifier.
                spectrum_dict['params']['title'] = \
                    f'{filename}:scan:{spectrum_dict["params"]["scans"]}'
            elif 'scan' in spectrum_dict['params']:
                spectrum_dict['params']['title'] = \
                    f'{filename}:scan:{spectrum_dict["params"]["scan"]}'
            else:
                # Use the index in the MGF file as identifier.
                spectrum_dict['params']['title'] = \
                    f'{filename}:index:{spectrum_i}'
            try:
                yield _parse_spectrum(spectrum_dict)
            except (ValueError, KeyError):
                pass


def _parse_spectrum(spectrum_dict: Dict) -> sus.MsmsSpectrum:
    """
    Parse the Pyteomics cluster dict.

    Parameters
    ----------
    spectrum_dict : Dict
        The Pyteomics cluster dict to be parsed.

    Returns
    -------
    MsmsSpectrum
        The parsed cluster.
    """
    identifier = spectrum_dict['params']['title']

    mz_array = spectrum_dict['m/z array']
    intensity_array = spectrum_dict['intensity array']
    retention_time = float(spectrum_dict['params'].get('rtinseconds', -1))

    precursor_mz = float(spectrum_dict['params']['pepmass'][0])
    if 'charge' in spectrum_dict['params']:
        precursor_charge = int(spectrum_dict['params']['charge'][0])
    else:
        raise ValueError('Unknown precursor charge')

    return sus.MsmsSpectrum(identifier, precursor_mz, precursor_charge,
                            mz_array, intensity_array, None, retention_time)


def write_spectra(filename: str, spectra: Iterable[sus.MsmsSpectrum]) -> None:
    """
    Write the given spectra to an MGF file.

    Parameters
    ----------
    filename : str
        The MGF file name where the spectra will be written.
    spectra : Iterable[MsmsSpectrum]
        The spectra to be written to the MGF file.
    """
    with open(filename, 'w') as f_out:
        pyteomics.mgf.write(_spectra_to_dicts(spectra), f_out, use_numpy=True)


def _spectra_to_dicts(spectra: Iterable[sus.MsmsSpectrum]) -> Iterable[Dict]:
    """
    Convert MsmsSpectrum objects to Pyteomics MGF cluster dictionaries.

    Parameters
    ----------
    spectra : Iterable[MsmsSpectrum]
        The spectra to be converted to Pyteomics MGF dictionaries.

    Returns
    -------
    Iterable[Dict]
        The given spectra as Pyteomics MGF dictionaries.
    """
    for spectrum in spectra:
        params = {'title': spectrum.identifier,
                  'pepmass': spectrum.precursor_mz,
                  'charge': spectrum.precursor_charge}
        if hasattr(spectrum, 'retention_time'):
            params['rtinseconds'] = spectrum.retention_time
        if hasattr(spectrum, 'scan'):
            params['scan'] = spectrum.scan
        if hasattr(spectrum, 'cluster'):
            params['cluster'] = spectrum.cluster
        yield {'params': params,
               'm/z array': spectrum.mz,
               'intensity array': spectrum.intensity}



pkg_name = 'falcon-ms'

try:
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version(pkg_name)
    except PackageNotFoundError:
        pass
except ImportError:
    from pkg_resources import get_distribution, DistributionNotFound

    try:
        __version__ = get_distribution(pkg_name).version
    except DistributionNotFound:
        pass
